import numpy as np
from gymnasium import spaces
import ezdxf
from openseespy.opensees import *
import os
import sys
from contextlib import contextmanager


class Cable:
    def __init__(self, obs_dim):
        self.obs_dim = obs_dim
        # 索力限制
        self.min_stress = 300
        self.max_stress = 1000
        self.min_num_strands = 5
        self.max_num_strands = 100
        self.stress_init = 0
        self.num_strands = 0
        self.stress_after = np.nan
        self.deform = None
        self.action = (None, None)
        self._reward = None

    @property
    def observation_space(self):
        return spaces.Box(
            low=np.float32(-10),
            high=np.float32(10),
            shape=(self.obs_dim,),
            dtype=np.float32,
        )

    @property
    def action_space(self):
        stress_act_space = spaces.Box(
            low=np.float32(0),
            high=np.float32(1),
            shape=(1,),
            dtype=np.float32,
        )
        strands_act_space = spaces.Discrete(int(self.max_num_strands - self.min_num_strands + 1))
        return spaces.Tuple([stress_act_space, strands_act_space])

    def reset(self):
        action = self.action_space.sample()
        self.stress_init = np.float32(self.min_stress + action[0][0] * (self.max_stress - self.min_stress))
        self.num_strands = self.min_num_strands + action[1]
        self.stress_after = np.nan
        self.action = action
        self.deform = None

    def step(self, norm_action):
        self.stress_init = np.float32(self.min_stress + norm_action[0] * (self.max_stress - self.min_stress))
        self.num_strands = self.min_num_strands + norm_action[1]
        return

    def update(self, balance_stress, deform):
        self.stress_after = balance_stress
        self.deform = deform

    def done(self):
        return self.num_strands <= 0 or self.stress_init <= 0 or self.stress_init >= 2000

    def reward2(self):
        EPS = 0.01
        if abs(self.deform) <= EPS:
            s_score = {0: -1, 1: 2, 2: -1}
        else:
            if self.deform > 0:
                s_score = {0: 2, 1: -1, 2: -1}
            else:
                s_score = {0: -1, 1: -1, 2: 2}
        if self.stress_after <= 300:
            n_score = {0: 1, 1: 0, 2: -1}
        elif self.stress_after <= 800:
            n_score = {0: 0, 1: 1, 2: 0}
        else:
            n_score = {0: -2, 1: -1, 2: +2}
        self._reward = s_score[self.action[0]] + n_score[self.action[1]]
        return self._reward

    def reward3(self):
        self._reward = -abs(self.deform) * 100
        return self._reward


@contextmanager
def suppress_openseespy_output():
    with open(os.devnull, 'w') as devnull:
        original_stdout = sys.stdout
        original_stderr = sys.stderr
        try:
            sys.stdout = devnull
            sys.stderr = devnull
            yield
        finally:
            sys.stdout = original_stdout
            sys.stderr = original_stderr


class FEM:
    def __init__(self, cables_per_side: int, Wg: float, tensions):
        self.nodes = []  # 节点坐标信息
        self.elements = []  # 单元连接节点信息
        self.element_types = {}  # 单元类型信息
        self.materials = {}  # 单元的材料信息
        self.cab_per_side = cables_per_side  # 单侧索对
        self.Wg = Wg  # 梁单元线荷载 N/m
        self.tensions = tensions

    def add_node(self, node_id, x, y):
        self.nodes.append({'id': node_id, 'x': x, 'y': y})

    def add_element(self, element_id, node1_id, node2_id, element_type, material_id):
        self.elements.append({'id': element_id, 'node1': node1_id, 'node2': node2_id, 'type': element_type, 'material': material_id})

    def add_element_type(self, type_id, description):
        self.element_types[type_id] = description

    def add_material(self, material_id, properties):
        self.materials[material_id] = properties

    def generate_dxf(self, filename):
        """
        生成 DXF 文件
        :param filename: 要保存的 DXF 文件名
        """
        # 创建一个新的 DXF 文档
        doc = ezdxf.new(dxfversion='R2010')
        # 添加一个新的模型空间
        msp = doc.modelspace()
        # 定义颜色映射
        color_map = {
            1: 1,  # 红色
            2: 2,  # 黄色
            3: 3,  # 绿色
            4: 4,  # 青色
            5: 5,  # 蓝色
            6: 6,  # 紫色
        }
        for i in range(100):
            color_map[1001 + i] = 100 + i
        # 添加节点
        for n in self.nodes:
            msp.add_circle(center=(n['x'], n['y']), radius=0.1, dxfattribs={'color': 7})
        # 添加单元
        for e in self.elements:
            node1 = next(n for n in self.nodes if n['id'] == e['node1'])
            node2 = next(n for n in self.nodes if n['id'] == e['node2'])
            color = color_map.get(e['material'], 1)  # 默认颜色为红色
            msp.add_line(start=(node1['x'], node1['y']), end=(node2['x'], node2['y']), dxfattribs={'color': color})
        # 保存 DXF 文件
        doc.saveas(filename)

    def opensees(self):
        """
        计算
        """
        with suppress_openseespy_output():
            wipe()
            model('basic', '-ndm', 2, '-ndf', 3)
            for nd in self.nodes:
                node(nd['id'], nd['x'], nd['y'])
                if nd['id'] > 1000:
                    fix(nd['id'], 1, 1, 1)
            fix(1, 0, 1, 0)  # 桥台
            fix(self.cab_per_side // 2 + 2, 0, 1, 0)  # 索塔
            fix(self.cab_per_side + 3, 1, 0, 1)  # 跨中
            fix(self.cab_per_side // 2 * 3 + 4, 0, 1, 0)  # 索塔
            fix(self.cab_per_side * 2 + 5, 0, 1, 0)  # 桥台
            geomTransf('Linear', 1)
            A = self.materials[1]['A']
            E = self.materials[1]['E']
            Iz = self.materials[1]['I']
            # 定义拉索属性
            Es = 1.95e11  # 弹性模量，单位：Pa (N/m²)
            As = 0.00014
            uniaxialMaterial('Elastic', 2, Es)
            for ed in self.elements:
                if ed['type'] == 1:
                    element('elasticBeamColumn', ed['id'], ed['node1'], ed['node2'], A, E, Iz, 1)
                else:
                    # continue
                    n1 = next(n for n in self.nodes if n['id'] == ed['node1'])
                    mat = self.materials[n1['id'] % 1000 + 1000]
                    uniaxialMaterial('InitStressMaterial', ed['id'], 2, float(mat['sigma'] * 1e6))
                    area = As * mat['Ns']  # 截面面积，单位：m²
                    element('corotTruss', ed['id'], ed['node1'], ed['node2'], area, ed['id'])
            # 设置均布荷载
            timeSeries('Linear', 1)
            pattern('Plain', 1, 1)
            for i in range(1, self.cab_per_side * 2 + 4):
                eleLoad('-ele', i, '-type', '-beamUniform', -self.Wg)
            system('BandSPD')
            constraints('Plain')
            numberer('Plain')
            # test('NormDispIncr', 1.0e-3, 100, 0)
            test('NormUnbalance', 1.0e-6, 100)
            integrator('LoadControl', 1)
            algorithm("Linear")
            # algorithm("Newton")
            analysis("Static")
            # printModel()
            analyze(1)
            res = []
            e_res = []
            for nd in self.nodes:
                if nd['id'] < 1000:
                    res.append(nodeDisp(nd['id'])[1])
            for i in range(self.cab_per_side // 2):
                eid = 1001 + i
                fx, fy = eleForce(eid)[0:2]
                mat = self.materials[eid]
                sig = (fx ** 2 + fy ** 2) ** 0.5 / (mat['Ns'] * As)
                e_res.append(sig * 1e-6)
            for i in range(self.cab_per_side // 2):
                eid = 2000 + self.cab_per_side // 2 - i
                fx, fy = eleForce(eid)[0:2]
                mat = self.materials[eid]
                sig = (fx ** 2 + fy ** 2) ** 0.5 / (mat['Ns'] * As)
                e_res.append(sig * 1e-6)
            wipe()
            # print(len(res))
            return res, e_res



import copy
from typing import Any

import gymnasium as gym
import numpy as np
from ding.envs import BaseEnv, BaseEnvTimestep
from ding.torch_utils import to_ndarray
from ding.utils import ENV_REGISTRY
from easydict import EasyDict


@ENV_REGISTRY.register('cable_bridge_dix')
class CableBridgeDi(BaseEnv):
    config = dict(
        env_id='cable_bridge_di-v0',
        act_scale=True,
    )
    default_env_id = ['cable_bridge_di-v0', ]

    @classmethod
    def default_config(cls: type) -> EasyDict:
        cfg = EasyDict(copy.deepcopy(cls.config))
        cfg.cfg_type = cls.__name__ + 'Dict'
        return cfg

    def __init__(self, cfg: EasyDict) -> None:
        self._cfg = cfg
        self._env_id = cfg.env_id
        assert self._env_id in self.default_env_id
        self._act_scale = cfg.act_scale
        self._replay_path = None
        self._save_replay = False
        self._save_replay_count = 0
        self._init_flag = False

    def reset(self) -> np.ndarray:
        if not self._init_flag:
            self._env = gym.make(self._env_id)
            self._observation_space = self._env.observation_space
            self._action_space = self._env.action_space
            self._reward_space = gym.spaces.Box(low=-np.inf, high=np.inf, shape=(1,), dtype=np.float32)
            self._init_flag = True
        self._env.seed(0)
        self._eval_episode_return = 0
        obs, _ = self._env.reset()
        obs = to_ndarray(obs).astype(np.float32)
        return obs

    def close(self) -> None:
        if self._init_flag:
            self._env.close()
        self._init_flag = False

    def step(self, action: Any) -> 'BaseEnv.timestep':
        obs, rew, done, trunc, info = self._env.step(action)
        obs = to_ndarray(obs)
        obs = obs.astype(np.float32)
        rew = to_ndarray([rew])  # wrapped to be transferred to a numpy array with shape (1,)
        if isinstance(rew, list):
            rew = rew[0]
        assert isinstance(rew, np.ndarray) and rew.shape == (1,)
        self._eval_episode_return += rew.item()
        return BaseEnvTimestep(obs, rew, done, info)

    def seed(self, seed: int) -> None:
        self._seed = seed
        np.random.seed(self._seed)

    @property
    def observation_space(self) -> gym.spaces.Space:
        return self._observation_space

    @property
    def action_space(self) -> gym.spaces.Space:
        return self._action_space

    @property
    def reward_space(self) -> gym.spaces.Space:
        return self._reward_space

    def __repr__(self) -> str:
        pass



from gymnasium.envs.registration import register


register(
    id='cable_bridge_di-v0',
    entry_point='bridgezoo.cablebridge_di.envs.cablebridge_base:CableBridgeBase',
    max_episode_steps=200,
)



from collections import OrderedDict

import gymnasium as gym
import numpy as np
import pygame

from ezdxf.math import Matrix44, Vec2
from gymnasium import spaces
from gymnasium.utils import seeding

from .cablebridge_models import FEM, Cable


class CableBridgeBase(gym.Env):
    metadata = {"render_modes": ["human", "rgb_array"], "render_fps": 30}

    def __init__(
            self,
            beam_e=30e9,
            beam_w=10.0,
            beam_h=1.0,
            num_cables_per_side=30,
            anchor_height=80,
            max_cycles=10,
            render_mode=None,
            DEF_SCALE=10,
            FPS=10,
    ):

        self.clock = pygame.time.Clock()
        self.DEF_SCALE = DEF_SCALE
        self.max_cycles = max_cycles
        # 物理参数 Start --------------------------------------------------
        w = beam_w
        h = beam_h
        self.beam_area = h * w
        self.beam_Iz = w * h ** 3 / 12.0
        self.beam_E = beam_e
        self.wg = w * h * 1 * 2400 * 9.806
        self.num_cables_per_side = num_cables_per_side
        self.middle_spacing = 10
        self.outside_spacing = 8
        self.end_to_first_spacing = 4
        self.center_to_adjacent_spacing = 2
        self.vertical_spacing = 2
        self.anchor_height = anchor_height  # 上塔柱高度
        self.num_beam_points = self.num_cables_per_side + 3  # 只考虑一侧的梁节点
        self.span = 2 * (self.num_cables_per_side * 0.5 * self.middle_spacing + self.center_to_adjacent_spacing)
        self.side_span = self.end_to_first_spacing + self.num_cables_per_side * 0.5 * self.outside_spacing
        self.beam_length = self.side_span * 2 + self.span
        self.x_positions = np.zeros(self.num_cables_per_side * 2 + 5)
        self.left_tower_top = self.right_tower_top = Vec2(0, 0)
        self.left_tower_base = self.right_tower_base = Vec2(0, 0)
        self.left_tower_pts = []
        self.right_tower_pts = []
        # 主梁参数
        x1 = np.linspace(-0.5 * self.beam_length + self.end_to_first_spacing, -self.outside_spacing - self.span * 0.5, self.num_cables_per_side // 2)
        x2 = np.linspace(-self.span * 0.5, -self.center_to_adjacent_spacing, self.num_cables_per_side // 2 + 1)
        self.x_positions = np.hstack((-self.beam_length * 0.5, x1, x2, 0, x2 * -1, x1 * -1, 0.5 * self.beam_length))
        self.x_positions.sort()
        # 索塔参数
        self.left_tower_top = Vec2([-0.5 * self.span, self.anchor_height])
        self.right_tower_top = Vec2([0.5 * self.span, self.anchor_height])
        self.left_tower_base = Vec2(-0.5 * self.span, 0)
        self.right_tower_base = Vec2(0.5 * self.span, 0)
        self.left_tower_pts = []
        self.right_tower_pts = []
        for i in range(self.num_cables_per_side // 2):
            left_anchor = self.left_tower_top + Vec2(0, -self.vertical_spacing) * i
            right_anchor = self.right_tower_top + Vec2(0, -self.vertical_spacing) * i
            self.left_tower_pts.append(left_anchor)
            self.right_tower_pts.append(right_anchor)

        # 物理参数 End  --------------------------------------------------
        obv_dim = self.num_cables_per_side + 1
        self.cables = OrderedDict({"Cable_%i" % i: Cable(obv_dim) for i in range(self.num_cables_per_side)})
        self.render_mode = render_mode
        self.screen = None
        self.frames = 0
        self.num_agents = self.num_cables_per_side
        self.observation_space = spaces.Box(low=-np.inf, high=np.inf, shape=(self.num_cables_per_side * 2,), dtype=np.float32)
        act_space = [c.action_space for i, c in self.cables.items()]
        act_n = self.cables['Cable_0'].action_space[1].n
        self.action_space = spaces.Dict({
            "discrete_actions": spaces.MultiDiscrete([act_n, ] * self.num_cables_per_side),  # 4组离散动作，每组有4个值
            "continuous_actions": spaces.Box(low=0, high=1.0, shape=(self.num_cables_per_side,), dtype=np.float32)  # 2个连续动作，取值范围[-1, 1]
        })
        # self.action_space = spaces.Tuple(act_space)
        self.screen_width = 1080
        self.screen_height = 600
        self.FPS = FPS
        self.screen_scale = (self.screen_width - 100) / self.beam_length  # 视频比例尺
        self.seed()
        self.state = None

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def _seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def close(self):
        if self.screen is not None:
            pygame.quit()
            self.screen = None

    def reset(self, **kwargs):
        self.frames = 0
        self.cables = OrderedDict({"Cable_" + str(r): Cable(self.num_cables_per_side + 1) for r in range(self.num_cables_per_side)})
        for i, c in self.cables.items():
            c.reset()
        cable_stress = [c.stress_init for i, c in self.cables.items()]
        cable_no = [c.num_strands for i, c in self.cables.items()]
        beam_pos, cable_stress_after = self.update_fem(cable_stress, cable_no)
        beam_pos = beam_pos[:self.num_cables_per_side + 3]
        beam_pos = beam_pos[1:self.num_cables_per_side // 2 + 1] + beam_pos[self.num_cables_per_side // 2 + 1 + 1:-1]
        for i, (key, c) in enumerate(self.cables.items()):
            c.update(cable_stress_after[i], beam_pos[i])
        self.state = np.hstack((beam_pos, cable_stress_after), dtype=np.float32)
        return self.state, {}

    def step(self, action_dict):
        dis_act = action_dict['discrete_actions']
        con_act = action_dict['continuous_actions']
        for i, (k, c) in enumerate(self.cables.items()):
            self.cables[k].step((con_act[i], dis_act[i]))

        cable_stress = [c.stress_init for i, c in self.cables.items()]
        cable_no = [c.num_strands for i, c in self.cables.items()]
        beam_pos, cable_stress_after = self.update_fem(cable_stress, cable_no)
        beam_pos = beam_pos[:self.num_cables_per_side + 3]
        beam_pos = beam_pos[1:self.num_cables_per_side // 2 + 1] + beam_pos[self.num_cables_per_side // 2 + 1 + 1:-1]
        for i, (key, c) in enumerate(self.cables.items()):
            c.update(cable_stress_after[i], beam_pos[i])
        self.state = np.hstack((beam_pos, cable_stress_after), dtype=np.float32)
        rewards = [c.reward3() for i, c in self.cables.items()]
        done = False
        truncated = self.frames >= self.max_cycles
        self.frames += 1
        if self.render_mode == "human":
            self.render()
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    pygame.quit()
                    exit()
        return self.state, np.sum(rewards), done, truncated, {}

    def render(self):
        if self.render_mode is None:
            gym.logger.warn(
                "You are calling render method without specifying any render mode."
            )
            return

        if self.screen is None:
            if self.render_mode == "human":
                pygame.init()
                self.screen_width = 1080
                self.screen_height = 600
                self.screen_scale = (self.screen_width - 100) / self.beam_length  # 视频比例尺
                self.screen = pygame.display.set_mode((self.screen_width, self.screen_height))
                pygame.display.set_caption('Cable-Stayed Bridge Environment')

        beam_positions = np.hstack(
            (
                [0, ],
                self.state[:self.num_cables_per_side // 2],
                [0, ],
                self.state[self.num_cables_per_side // 2:self.num_cables_per_side],
                self.state[self.num_cables_per_side - 1],
            )
        )
        cable_stress_after = [cable.stress_after for i, cable in self.cables.items()]
        if not hasattr(self, 'screen'):
            pygame.init()
            self.screen_width = 1080
            self.screen_height = 600
            self.screen_scale = (self.screen_width - 100) / self.beam_length  # 视频比例尺
            self.screen = pygame.display.set_mode((self.screen_width, self.screen_height))
            pygame.display.set_caption('Cable-Stayed Bridge Environment')
        self.screen.fill((255, 255, 255))  # 清空屏幕并设置背景为白色
        width, height = self.screen.get_size()
        screen_center = (width // 2, int(height * 3 / 4))
        mat = self.create_transformation_matrix(self.screen_scale, self.screen_scale, 0, screen_center[0], screen_center[1])

        # 绘制淡淡的网格线，水平间距为50米
        center_x, center_y = width // 2, int(height * 3 / 4)
        for i in range(center_x % 50, width, 50):
            pygame.draw.line(self.screen, (200, 200, 200), (i, 0), (i, height))
        for j in range(center_y % 50, height, 50):
            pygame.draw.line(self.screen, (200, 200, 200), (0, j), (width, j))

        # 绘制梁的位置
        y_positions = np.hstack((beam_positions, beam_positions[::-1][1:]))
        transformed_points = self.trans(mat, list(zip(self.x_positions, y_positions * self.DEF_SCALE)))
        pygame.draw.lines(self.screen, (0, 0, 0), False, transformed_points, 5)
        for pt in transformed_points:
            pygame.draw.line(self.screen, (255, 0, 0), Vec2(pt) + Vec2(0, 10), Vec2(pt) + Vec2(0, -10), 1)

        # 绘制固定点
        pygame.draw.circle(self.screen, (0, 0, 0), transformed_points[0], 5)
        pygame.draw.circle(self.screen, (0, 0, 0), transformed_points[-1], 5)

        # 计算左侧和右侧拉索锚点基准位置
        for i in range(self.num_cables_per_side // 2):
            left_anchor = self.left_tower_pts[i]
            right_anchor = self.right_tower_pts[i]
            tr2 = self.trans(mat, (left_anchor, right_anchor))
            beam_index_left = i + 1
            beam_index_right = self.num_beam_points
            pygame.draw.line(self.screen, (0, 0, 255), tr2[0], transformed_points[beam_index_left], 2)
            pygame.draw.line(self.screen, (0, 0, 255), tr2[0], transformed_points[self.num_beam_points - 2 - i], 2)
            pygame.draw.line(self.screen, (0, 0, 255), tr2[1], transformed_points[beam_index_right + i], 2)
            pygame.draw.line(self.screen, (0, 0, 255), tr2[1], transformed_points[self.num_beam_points * 2 - 3 - i], 2)

        # 绘制参考线
        left_anchor_base = self.trans(mat, self.left_tower_top)
        right_anchor_base = self.trans(mat, self.right_tower_top)
        left_tower_base = self.trans(mat, self.left_tower_base)
        right_tower_base = self.trans(mat, self.right_tower_base)
        pygame.draw.line(self.screen, (0, 0, 0), left_tower_base, left_anchor_base, 4)
        pygame.draw.line(self.screen, (0, 0, 0), right_tower_base, right_anchor_base, 4)

        # 显示梁的总长度、固定点高度和网格间距
        font = pygame.font.SysFont('fs', 20)
        text = font.render('Tower Height: %.1f m' % self.anchor_height, True, (0, 0, 0))
        self.screen.blit(text, (10, 10))
        text = font.render('Span Length:%.1f m' % self.span, True, (0, 0, 0))
        self.screen.blit(text, (10, 25))
        text = font.render(f'EI: {self.beam_area}', True, (0, 0, 0))
        self.screen.blit(text, (10, 40))
        text = font.render('Beam0Y: %.0f mm' % (float(beam_positions[-1]) * 1000), True, (0, 0, 0))
        self.screen.blit(text, (10, 55))
        text = font.render('Stress: %.1f MPa(Max. ) | %.1f MPa(Min.)' % (max(cable_stress_after), min(cable_stress_after)), True, (0, 0, 0))
        self.screen.blit(text, (10, 70))
        pygame.display.flip()
        self.clock.tick(self.FPS)

        if self.render_mode == "human":
            pygame.event.pump()
            pygame.display.update()
        return

    @staticmethod
    def create_transformation_matrix(sx, sy, angle, tx, ty):
        """
        创建一个二维的旋转、平移和缩放变换矩阵。

        :param sx: x方向的缩放因子
        :param sy: y方向的缩放因子
        :param angle: 逆时针旋转角度（以弧度表示）
        :param tx: x方向的平移量
        :param ty: y方向的平移量
        :return: 一个表示旋转、平移和缩放的组合变换矩阵
        """
        # 创建旋转矩阵（绕z轴逆时针旋转）
        rotation_matrix = Matrix44.z_rotate(angle)

        # 创建平移矩阵
        translation_matrix = Matrix44.translate(tx, ty, 0)

        # 创建缩放矩阵
        scaling_matrix = Matrix44.scale(sx, sy, 1)

        # 组合变换矩阵（先旋转，再平移，最后缩放）
        transformation_matrix = scaling_matrix @ translation_matrix @ rotation_matrix
        return transformation_matrix

    @staticmethod
    def trans(matrix, points):
        """
        应用变换矩阵到点集合上，返回变换后的点。

        :param matrix: 变换矩阵
        :param points: 点的集合，每个点为一个元组(x, y)
        :return: 变换后的点的集合
        """
        # 将点转换为 Vec2 对象并应用变换矩阵,上下反转
        if not isinstance(points, Vec2):
            transformed_points = [matrix.transform(Vec2(p[0], -p[1])) for p in points]
            return [(p.x, p.y) for p in transformed_points]
        else:
            transformed_points = matrix.transform(Vec2(points.x, -points.y))
            return transformed_points.vec2

    def extract_fem_model(self, cable_sigma, cable_sizes):
        """
        从CableSystemEnv提取信息，生成FEM实例
        """
        # cable_tensions = self.state[self.num_beam_points:self.num_beam_points + self.num_cables_per_side]
        fem_model = FEM(self.num_cables_per_side, self.wg, cable_sigma)
        # 提取梁的节点信息
        for i, x in enumerate(self.x_positions):
            fem_model.add_node(i + 1, x, 0)
        # 提取梁的单元信息
        for i in range(len(self.x_positions) - 1):
            fem_model.add_element(i + 1, i + 1, i + 2, 1, 1)  # 梁单元类型和材料假设为1
        # 提取索塔坐标
        for i in range(self.num_cables_per_side // 2):
            fem_model.add_node(1001 + i, self.left_tower_pts[i].x, self.left_tower_pts[i].y)
            fem_model.add_node(3001 + i, self.right_tower_pts[i].x, self.right_tower_pts[i].y)
            beam_index_left = i + 2
            beam_index_right = self.num_beam_points + 1
            fem_model.add_element(1001 + i, 1001 + i, beam_index_left, 2, 1001 + i)  #
            fem_model.add_element(2001 + i, 1001 + i, self.num_beam_points - 1 - i, 2, 1001 + i)
            fem_model.add_element(3001 + i, 3001 + i, beam_index_right + i, 2, 1001 + i)
            fem_model.add_element(4001 + i, 3001 + i, self.num_beam_points * 2 - 2 - i, 2, 1001 + i)
            Ns1001 = cable_sizes[i]
            Nf1001 = cable_sigma[i]
            Ns2001 = cable_sizes[self.num_cables_per_side - 1 - i]
            Nf2001 = cable_sigma[self.num_cables_per_side - 1 - i]
            fem_model.add_material(1001 + i, {'Ns': Ns1001, 'sigma': Nf1001})
            fem_model.add_material(2001 + i, {'Ns': Ns2001, 'sigma': Nf2001})
            fem_model.add_material(3001 + i, {'Ns': Ns1001, 'sigma': Nf1001})
            fem_model.add_material(4001 + i, {'Ns': Ns2001, 'sigma': Nf2001})

        # 添加单元类型和材料信息
        fem_model.add_element_type(1, 'Beam')
        fem_model.add_element_type(2, 'Cable')
        fem_model.add_material(1, {'E': self.beam_E, 'A': self.beam_area, 'I': self.beam_Iz})

        return fem_model

    def update_fem(self, cable_force, cable_soq):
        fem = self.extract_fem_model(cable_force, cable_soq)
        return fem.opensees()
